<?php
require_once('AdministrateCredentials.php');
require_once('AdministrateCallValidator.php');

// Extend SoapClient so that we can catch the response mid-request and mutate it.
class AdministrateSoapClient extends \SoapClient{

	private function _strip_control_characters($input) {
		return preg_replace('/[\x00-\x1F\x7F]/', '', $input);
	}

	public function __doRequest($req, $location, $action, $version=SOAP_1_1, $one_way=0){
		$raw_xml = parent::__doRequest($req, $location, $action, $version, $one_way);
		return $this->_strip_control_characters($raw_xml);
	}
}

//  Administrate raw API
//
class AdministrateBaseAPI {
	private $_api_credentials = null;

	/* @var $_soap_client AdministrateSoapClient */
	private $_soap_client = null;
	private $_debug = null;
	private $_rest_url = null;
	private $_soap_url = null;
	private $_internal_api_url = null;
	private $_has_curl = null;
	private $_user = null;
	private $_password = null;

	//  Setup API connection
	public function set_credentials($api_credentials) {
		if($this->_api_credentials) {
			$this->_warn("API credentials already set.");
			return;
		}

		$this->_api_credentials = $api_credentials;

		$this->_debug = $api_credentials->get_debug();

		$this->_internal_api_url = $api_credentials->get_internal_api_url();
		$this->_rest_url	= $api_credentials->get_rest_url();
		$this->_has_curl 	= false;

		$this->_soap_url	= $api_credentials->get_soap_url();
		$this->_soap_client = null;

		$this->_user			= $api_credentials->get_user();
		$this->_password	= $api_credentials->get_password();
		
		// Assert credentials
		if( $api_credentials->is_empty() ) {
			$this->_warn("API credentials not set.");
		}
		else
		{
			// Initialize API
			$this->_init_rest();
			$this->_init_soap();
		}
	}
	
	public function get_credentials() {
		return $this->_api_credentials;
	}
	
	//  Initialize REST API
	protected function _init_rest() {
		
		if( !function_exists("curl_init")) {
			$this->_log("CURL not installed.");
			return false;
		}
		
		$this->_has_curl = true;
		$this->_log("REST initialized.");
		
		return true;
	}
	
	//  Initialize SOAP API
	protected function _init_soap() {
		$this->_log("SOAP initialization request.");

		try {
			if( !class_exists("AdministrateSoapClient") ) {
				throw new Exception("SOAP not installed.");
			}

			//  Disable soap cache
			ini_set('soap.wsdl_cache_enabled', 0);

			//  Initialize soap client
			$this->_soap_client = new AdministrateSoapClient(
				$this->_soap_url,
				array(
					'trace'			=>	$this->_debug,
					'exception'	=>	1,
					'login'			=>	$this->_user,
					'password'	=>	$this->_password,
					'features'	=>	SOAP_SINGLE_ELEMENT_ARRAYS
				)
			);
			$this->_log('SOAP initialized, url=' . $this->_soap_url);
			return true;
			
		} catch(Exception $e) {
				$this->_error('SOAP initialization error: '.$e->getMessage());
				$this->_soap_client = null;
				return false;
		}
	}
	
	/*
		Raw API
	*/
	
	//  Raw API call
	public function raw_call($method, $args=array()) {
		try {
			$t_start = microtime(true);
			$this->_log('raw_call: ' . $method . ' [' . var_export($args, true) . ']');
			// Assert credentials
			if( is_null($this->_api_credentials) || $this->_api_credentials->is_empty() ) {
				throw new Exception("API credentials not set.");
			}
			
			// Currently only get events is performed as REST call
			if( $this->_rest_call_vaildate_args($method, $args) ) {
				if($method == 'getEvents') {
					if (!($events = $this->_rest_call($method, $args))) {
						$events = array();
					}

					// Return object in SOAP format
					$call_result = new stdClass();
					$call_result->EventList = $events;
				} else {
					return $this->_rest_call($method, $args);
				}
			} 
			// SOAP call
			else {
				if( !($call_result=$this->_soap_call($method, $args)) ) {
					return false;
				}
			}
		
			$t_end = microtime(true);
			$this->_log('Ok, exec_time='. number_format($t_end-$t_start,5) );
			return $call_result;
		}
		// Catch API exceptions 
		catch(Exception $e) {
			$this->_warn('API validation error: '.$e->getMessage());
			return false;
		}
		
	}
	
	// SOAP Call
	protected function _soap_call($method, $args) {

		$this->_log('SOAP call');
		
		// Assert SOAP has been initialized
		if( is_null($this->_soap_client) ) {
			$this->_error('Error making SOAP call: Not initialzied.');
			return false;
		}

		// If the arguments aren't already an array, put the value into an array
		if (!is_array($args)) {
			$args = array($args);	
		}
		
		// Make SOAP call
		try {
			$call_result = $this->_soap_client->__soapCall($method, $args);
			if(!$call_result) {
				return false;
			}
			
			// If the SOAP call has returned a CategoryList, the ordering of objects is the one specified
			// in the application and this has to be recored in cached objects.
			if( is_object($call_result) && property_exists($call_result, "CategoryList") ) {
				$i = 0;
				foreach ($call_result->CategoryList as $category) {
					$category->Index = $i++;
				}
			}
			return $call_result;
		} 
		catch (Exception $e) {
			$this->_error('Error making SOAP call: '.$e->getMessage());
			return false;
		}
	}
	
	// REST Call
	protected function _rest_call($method, $args) {
		$this->_log("REST call");

		if( !$this->_has_curl ) {
			$this->_log("REST not initialized (No Curl).");
			return array();
		}

		// Configure curl
		$ch = curl_init();

		if( $method == "getEvents" ) {
			$method = "events";
			$url = $this->_rest_url . $method;
			$url_args = array();

			if( array_key_exists('IncludeCurrentOnline', $args['Filter']) ) {
				$url_args['include_current_online'] = $args['Filter']['IncludeCurrentOnline'];
			}

			if( array_key_exists('IncludeSessions', $args['Filter']) ) {
				$url_args['include_sessions'] = $args['Filter']['IncludeSessions'];
			}

			if(array_key_exists('CourseCode', $args['Filter'])) {
				$url_args['course_code'] = $args['Filter']['CourseCode'];
			}

			if (count($url_args) > 0) {
				$query_string = "?". http_build_query($url_args);
				$url .= $query_string;
			}
		} elseif($method == 'place_opps_order') {
			$url = $this->_internal_api_url . 'order';
			$data = json_encode($args);
			curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
			curl_setopt($ch, CURLOPT_HTTPHEADER, array(
					'Content-Type: application/json',
					'Content-Length: ' . strlen($data))
			);
			curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
		} elseif($method == 'get_learning_tracks') {
			$url = $this->_internal_api_url . 'learning_tracks';
		} elseif($method == 'get_learning_track') {
			$url = $this->_internal_api_url . 'learning_tracks/' . $args;
		} elseif($method == 'get_mapped_items') {
			$url = $this->_internal_api_url . 'mapped_items/' . $args['course_id'] . '/' . $args['location_id'];
		} elseif($method == 'get_mapped_item') {
			$url = $this->_internal_api_url . 'mapped_items/' . $args;
		} elseif($method == 'getEventByID') {
			if(array_key_exists('eventID', $args)) {
				$event_id = $args['eventID'];
			} else {
				$event_id = $args[0];
			}
			$url = $this->_rest_url . 'events/' . $event_id;
		} else {
			curl_close($ch);
		}

		if($ch) {

			// Disable SSL verification
			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);

			// Will return the response, if false it prints the response
			curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
			curl_setopt($ch, CURLOPT_URL, $url);
			curl_setopt($ch, CURLOPT_USERPWD, $this->_user . ':' . $this->_password);
			curl_setopt($ch, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);

			$result = curl_exec($ch);
			curl_close($ch);
			return json_decode($result);
		} else {

			return null;
		}
	}

	// Validate REST call arguments
	private function _rest_call_vaildate_args( $method, $args ) {
		$valid_args = array(
			"IncludePrices",
			"IncludeCurrentOnline",
			"IncludeSessions",
			"CourseCode"
		);

		// These are currently the only REST methods - anything else is not a REST method therefor isn't valid.
		if(
			!in_array(
				$method,
				array(
					"getEvents",
					"getEventByID",
					"place_opps_order",
					"get_learning_tracks",
					"get_learning_track",
					"get_mapped_items",
					"get_mapped_item"
				)
			)
		) {
			return false;
		}

		if($method == 'getEvents') {
			if (!array_key_exists("Filter", $args)) {
				return false;
			}

			foreach ($args["Filter"] as $name => $value) {
				if (!in_array($name, $valid_args)) {
					return false;
				}
			}
			return true;
		} else {
			// @todo: Validate place_opps_order call
			return true;
		}
	}
	
	/*
		Internal event logging
	*/
	protected function _log($user_message) {
		Administrate::log()->notice("Administrate API: " . $user_message);
	}
	protected function _warn($user_message) {
		Administrate::log()->warning("Administrate API: " . $user_message);
	}
	protected function _error($user_message) {
		Administrate::log()->error("Administrate API: " . $user_message);
	}
	
}
