<?php
use PHPUnit\Framework\TestCase;
require_once('common/api/AdministrateCache.php');
require_once('weblink-components/config.php');

class AdministrateCacheTest extends TestCase {
	
	public function test_is_call_cacheable() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$this->assertTrue( $cache->is_call_cacheable("getEventByID") );
		$this->assertTrue( $cache->is_call_cacheable("getEvents") );
		$this->assertTrue( $cache->is_call_cacheable("getCourseByCode") );
		$this->assertTrue( $cache->is_call_cacheable("getCourses") );
		$this->assertTrue( $cache->is_call_cacheable("getCategories") );
		
		$this->assertFalse( $cache->is_call_cacheable("") );
		$this->assertFalse( $cache->is_call_cacheable("UnknownCall") );
		
		$this->assertFalse( $cache->is_call_cacheable("addDelegate") );
		$this->assertFalse( $cache->is_call_cacheable("sendInvoice") );
		$this->assertFalse( $cache->is_call_cacheable("getPaymentIntegrationConfig") );
		$this->assertFalse( $cache->is_call_cacheable("placeOrder") );
		$this->assertFalse( $cache->is_call_cacheable("getAllDiscounts") );
	}
	
	public function test_is_result_multiple() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$this->assertFalse( $cache->is_result_multiple("getEventByID") );
		$this->assertFalse( $cache->is_result_multiple("getCourseByCode") );
		
		$this->assertTrue( $cache->is_result_multiple("getEvents") );
		$this->assertTrue( $cache->is_result_multiple("getCourses") );
		$this->assertTrue( $cache->is_result_multiple("getCategories") );
	}
	
	public function test_get_call_type() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$type1 = $cache->get_call_type("getEventByID");
		$type2 = $cache->get_call_type("getEvents");
		$this->assertEquals($type1, $type2);
		
		$type1 = $cache->get_call_type("getCourseByCode");
		$type2 = $cache->get_call_type("getCourses");
		$this->assertEquals($type1, $type2);
		
		$type1 = $cache->get_call_type("getEvents");
		$type2 = $cache->get_call_type("getCourses");
		$type3 = $cache->get_call_type("getCategories");
		$this->assertNotEquals($type1, $type2);
		$this->assertNotEquals($type2, $type3);
		$this->assertNotEquals($type1, $type3);
	}
	
	public function test_fetch_single_object() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$obj = new StdClass();
		$obj->data = "Test Data";
		
		$cache_hit = $cache->fetch_single_object("getEventByID","id1");
		$this->assertEquals(null, $cache_hit);
		
		$cache->record_object("getEventByID", "id1", $obj);
		$cache_hit = $cache->fetch_single_object("getEventByID","id1");
		$this->assertEquals($obj, $cache_hit);
		
		$cache_hit = $cache->fetch_single_object("getEventByID","id2");
		$this->assertEquals(null,$cache_hit);
	}
	
	public function test_fetch_two_objects() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$obj1 = new StdClass();
		$obj1->data = "Test Data";
		
		$obj2 = new StdClass();
		$obj2->data = "Other Data";
		
		$cache->record_object("getEventByID", "id1", $obj1);
		$cache->record_object("getEventByID", "id1", $obj2);
		$cache_hit = $cache->fetch_single_object("getEventByID","id1");
		$this->assertEquals($obj2, $cache_hit);
		
		$db->truncate_tables();
		$cache->record_object("getEventByID", "id1", $obj1);
		$cache->record_object("getEventByID", "id2", $obj2);
		
		$cache_hit1 = $cache->fetch_single_object("getEventByID","id1");
		$cache_hit2 = $cache->fetch_single_object("getEventByID","id2");
		$this->assertEquals($obj1, $cache_hit1);
		$this->assertEquals($obj2, $cache_hit2);
	}
	
	public function test_fetch_multiple_objects() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$entity = $cache->get_call_type("getEvents");
		$entity_params = $cache->get_call_parameters($entity);
		
		$obj1 = new StdClass();
		$obj1->data = "Event";
		
		$obj2 = new StdClass();
		$obj2->data = "Other Event";
		
		$cache->record_object("getEventByID", "1", $obj1);
		$cache->record_object("getEventByID", "2", $obj2);
		
		$cache_hit = $cache->fetch_multiple_objects("getEvents","2",$entity_params);
		$this->assertNotEquals(null, $cache_hit);
		$this->assertObjectHasAttribute("EventList", $cache_hit);
		$this->assertEquals(1, count($cache_hit->EventList));
		$this->assertEquals($obj2, $cache_hit->EventList[0]);
		
		$cache_hit = $cache->fetch_multiple_objects("getEvents","2,1",$entity_params);
		$this->assertNotEquals(null, $cache_hit);
		$this->assertObjectHasAttribute("EventList", $cache_hit);
		$this->assertEquals(2, count($cache_hit->EventList));
		$this->assertEquals($obj1, $cache_hit->EventList[0]);
		$this->assertEquals($obj2, $cache_hit->EventList[1]);
	}
	
	public function test_fetch_objects_no_duplicates() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$entity = $cache->get_call_type("getEvents");
		$entity_params = $cache->get_call_parameters($entity);
		
		$obj = new StdClass();
		$obj->data = "Event";
		
		$cache->record_object("getEventByID", "objid1", $obj);
		$cache->record_object("getEventByID", "objid1", $obj);
		
		$cache_hit = $cache->fetch_multiple_objects("getEvents","objid1",$entity_params);
		$this->assertNotEquals(null, $cache_hit);
		$this->assertObjectHasAttribute("EventList", $cache_hit);
		$this->assertEquals(1, count($cache_hit->EventList));
		$this->assertEquals($obj, $cache_hit->EventList[0]);
	}
		
	public function test_fetch_api_call() {
		$db = new AdministrateCache_TestDB();
		$cache = new AdministrateCache($db,"","test");
		
		$status = $cache->record_api_call("getCourses", array(), array(1,2,3));
		$this->assertTrue($status);
		
		$cache_hit = $cache->fetch_api_call("getCourses", array());
		$this->assertEquals("1,2,3", $cache_hit);
	}
		
}

class AdministrateCache_TestDB {
	private $_connection = false;
	private $_last_result = 0;
	
	public function __construct() {
		// Connect to test db
		global $_CONFIG;
		$db_config = $_CONFIG["service"]["db"];
		
		$this->_connection = new mysqli($db_config["host"], $db_config["user"], $db_config["password"], $db_config["name"]);
		
		if( !$this->_connection ) {
			echo "Can't connect to test database\n";
			return;	
		}
		$this->truncate_tables();
	}
	
	public function truncate_tables() {
		// Truncate cache tables
		$this->_connection->query("TRUNCATE TABLE cached_calls");
		$this->_connection->query("TRUNCATE TABLE cached_categories");
		$this->_connection->query("TRUNCATE TABLE cached_courses");
		$this->_connection->query("TRUNCATE TABLE cached_events");
	}
	
	public function prepare($qry, $param_types = "") {
		$mysqli_pepared = $this->_connection->prepare($qry);
		
		if(!$this->_connection->prepare($qry)) {
			return false;
		}
		return array($param_types, $mysqli_pepared);
	}
	
	public function prepared_exec($prep_statement, $params = array()) {
		$param_types = $prep_statement[0];
		$mysqli_pepared = $prep_statement[1];

		$args_array = array( $param_types );
		
		for($i=0; $i<count($params); $i++) {
			$args_array[] = &$params[$i];
		}
		
		call_user_func_array(array($mysqli_pepared, "bind_param"), $args_array);
		$status = $mysqli_pepared->execute();
		$this->_last_result = $mysqli_pepared->get_result();
		return $status;
	}
	
	public function prepared_fetch($prep_statement, $params = array()) {
		$this->prepared_exec($prep_statement, $params);
		
		$mysqli_pepared = $prep_statement[1];
		
		if(!$this->_last_result) {
			return false;
		}
		
		$rows = $this->_last_result->fetch_all(MYSQLI_ASSOC);
		
		if( empty($rows) ) {
			return false;
		}
		
		return count($rows)==1 ? $rows[0] : $rows;
	}
	
	public function get_last_num_records() {
		if(!is_object($this->_last_result)) {
			return 0;
		}
		return $this->_last_result->num_rows;
	}
	
	public function format_string($val, $strip_slashes=true) {
		return $this->_connection->real_escape_string($strip_slashes ? stripslashes($val) : $val);
	}
	
}
