<?php
//  Data Access Object
class AdministrateDao {
	
	//  Properties
	protected $service;
	protected $widgetPath = '';
	protected $tableKey = '';
	protected $primaryKey = 'my_id';
	protected $fieldDefinitions = array();
	protected $row = false;
	
	//  Constructor
	public function __construct(&$service, $fields = array(), $where_on_api_key = true) {
		Administrate::log()->notice("DAO construct " . var_export($fields,true));
		
		//  Save the plugin by reference
		$this->service = $service;

		$reflector = new ReflectionClass(get_class($this));
		$this->widgetPath = dirname($reflector->getFileName());

		//  Set the field definitions
		$this->fieldDefinitions = $this->service->get_table_fields($this->tableKey);
		
		//  Save the actual table name using the key
		$this->table = $this->service->get_table($this->tableKey);
		
		//  If the passed ID is greater than zero, add it to the fields array
		if (!is_array($fields)) {
			$id = $fields;
			$fields = array();
			$fields[$this->_get_primary_key()] = $id;	
		}
		
		//  Try to get the database record based on fields to match against
		if (!empty($fields)) {

			// Add API Key field conditionally. We want to be able to explicitly ignore this check IF we're
			// looking up an order by ID, as we *know* these are going to be distinct per-client.
			if($where_on_api_key) {
				$fields[$this->apiKeyField] = $this->service->get_option('api_key');
			}

			$qry = 'SELECT * FROM ' . $this->_get_table() . ' WHERE ';
			foreach ($fields as $field=>$value) {
				$qry .= $field . " = '" . $value . "' AND ";	
			}
			$qry = substr($qry, 0, -5);
			//echo $qry . '<br>';
			$this->row = $this->service->db->query_fetch($qry);
			if (!is_array($this->row)) {
				$this->row = false;
			}
		}
		
	}

	//  Get the widget path
	public function get_path($path) {
		return $this->widgetPath.$path;
	}

	//  Save a record
	public function save($fields = array(), $force_insert = false) {
		Administrate::log()->notice("DAO save " . var_export($fields,true) );
		
		//  Only proceed if there were actually fields passed
		if (!empty($fields)) {
		
			//  Check the fields for errors
			$errors = $this->_validate_fields($fields);
			
			//  If there were no errors, proceed
			if (!$errors) {
				
				//  Sanitize the fields
				$fields = $this->_sanitize_fields($fields);
				
				//  Only proceed if there are still any fields to save after sanitation
				if (!empty($fields)) {
				
					//  If the row already exists, this is an update
					if ($this->row && !$force_insert) {
						
						//  Update the row
						$primary_key = $this->_get_primary_key();
						$this_key = $this->row[$primary_key];
						
						$this->service->db->update(
							$this->_get_table(),
							$fields,
							"{$primary_key} = {$this_key}"
						);
						
						//  Reload saved data
						$this->_reload_row();
						
					//  Otherwise this is an insert	
					} else {
					
						//  Add API key
						$fields[$this->apiKeyField] = $this->service->get_option('api_key');
					
						//  Insert the row
						$this->service->db->insert(
							$this->_get_table(),
							$fields
						);
						
						//  Reinitialize object
						$this->__construct($this->service, $this->service->db->get_last_insert_id());
							
					}
					
					return false;
				
				}
			
			//  Or else return the errors	
			} else {
				return $errors;
			}
		
		}
		
	}
	
	//  Create a new record
	// $dummy_field vars are because subclasses which we've overridden 'create' on can have more params
	public function create($fields = array(), $dummy_field=null, $dummy_field2=null) {
		return $this->save($fields);
	}
	
	//  Update an record
	public function update($fields = array()) {
		return $this->save($fields);
	}
	
	//  Delete a record
	public function delete() {
		Administrate::log()->notice("DAO delete");
		return $this->service->db->query("DELETE FROM " . $this->_get_table() . " WHERE " . $this->_get_primary_key() . " = " . $this->get_id());
	}
	
	//  Reload the row
	protected function _reload_row() {
		Administrate::log()->notice("DAO reload row");
		$this->row = $this->service->db->query_fetch('SELECT * FROM ' . $this->_get_table() . ' WHERE ' . $this->_get_primary_key() . ' = ' . $this->get_id(), 'ARRAY_A');	
	}
	
	//  Validate the fields -- By default does nothing, returns no errors
	protected function _validate_fields($fields = array()) { 
		return false;
	}
	
	//  Sanitize and pack up fields
	protected function _sanitize_fields($inputFields = array()) {
		Administrate::log()->notice("DAO sanitize fields ". var_export($inputFields,true));
		
		//  Loop through field definitions
		$outputFields = array();
		foreach ($this->fieldDefinitions as $field=>$properties) {
			if (isset($inputFields[$field])) {
				
				//  If this is a serialized field ...
				if (isset($properties['serialize'])) {
					
					//  Loop through serialized fields
					$outputFields[$field] = array();
					foreach ($properties['serialize'] as $subfield) {
						$subfieldKey = $field . $this->service->get_key_delimiter() . $subfield;
						if (is_array($inputFields[$subfieldKey])) {
							$outputFields[$field][$subfield] = array();
							foreach ($inputFields[$subfieldKey] as $multiField) {
								array_push($outputFields[$field][$subfield], $this->service->db->format_string($multiField),false);
							}
						} else {
							$outputFields[$field][$subfield] = $this->service->db->format_string($inputFields[$subfieldKey],false);
						}
					}
					//using base64 encode to deal with ' character in form data
					$outputFields[$field] = base64_encode(serialize($outputFields[$field]));
					
				//  Or else just sanitize the field
				} else {
					$outputFields[$field] = $this->service->db->format_string($inputFields[$field],false);
				}
				
			}
		}
		
		//  Return output fields
		return $outputFields;
		
	}
	
	//  Get the id
	public function get_id() {
		if ($this->row) {
			return $this->row[$this->_get_primary_key()];	
		} else {
			return 0;	
		}
	}
	
	//  Get table
	protected function _get_table() {
		return $this->table;	
	}
	
	//  Get the primary key
	protected function _get_primary_key() {
		return $this->primaryKey;	
	}
	
	//  Whether or not the passed ID exists
	public function exists() {
		return is_array($this->row);	
	}
	
	//  Get field
	public function get_field($key) {
		return $this->row[$key];	
	}
	
	//  Get all
	public function get_all($where = false, $order = false) {
		Administrate::log()->notice("DAO get all ". var_export($where));
		
		$qry = "SELECT * FROM " . $this->_get_table();
		if ($where) {
			$where[$this->apiKeyField] = $this->service->get_option('api_key');
			$qry .= " WHERE ";
			foreach ($where as $field=>$value) {
				$qry .= $field . " = '" . $value . "' AND ";	
			}
			$qry = substr($qry, 0, -5);
		}
		if ($order) {
			$qry .= " ORDER BY ";
			foreach ($order as $field=>$value) {
				$qry .= $field . " " . $value . ", ";	
			}
			$qry = substr($qry, 0, -2);
		}
		$rows = $this->service->db->query_fetch($qry);
		$class = get_class($this);
		$daos = array();
		foreach ($rows as $row) {
			array_push($daos, new $class($this->service, $row[$this->_get_primary_key()]));
		}
		return $daos;
	}
	
}
