<?php
// Install Weblink database schema
// Usage: 
// 1. backup:  mysqldump -u -p <weblink database> > weblink.sql
// 2. install: php weblink-components/db-install --db <weblink database> --user <username> --password <password>
//
require_once('common/administrate-common.php');
require_once('AdministrateMysqlDB.php');

// Get command line options
$cmd_options = getopt("", array("db:", "user:", "password:"));

if( !$cmd_options || 
		!array_key_exists("db", $cmd_options) || 
		!array_key_exists("user", $cmd_options) || 
		!array_key_exists("password", $cmd_options)  ) {
	
	print_info();
	exit();
}

// Connect to installation database
$weblink_db = new AdministrateMysqlDB(
	$cmd_options["db"],
	$cmd_options["user"],
	$cmd_options["password"],
	""
);
if( $weblink_db->get_last_error() != "" ) {
	echo "Aborting.\n";
	exit();
}

// Get API keys, if present
$api_keys = get_api_keys_v2($weblink_db);

if(!$api_keys) {
	$api_keys = get_api_keys_v1($weblink_db);
	
	if(!$api_keys) {
		echo "No API keys table.\n";
	}
}

// Install
echo "This tool will install the most recent Weblink database schema in '{$cmd_options['db']}'.\n";
echo "You should have backup of your database before running this tool.\n";
echo "Are you sure you want to install the new database schema? (yes/no) ";
if( user_confirmation() ) {
	
	db_install( $weblink_db );
	db_copy( $weblink_db, $api_keys );
	echo "Database schema installed.\n\n";
}
exit();

// Install database schema
function db_install($db) {
	
	// Cache tables
	$cache_tables = AdministrateCache::get_table_structure("");
	
	foreach($cache_tables as $tname => $tsql) {
		// Discard old cache table
		$db->query("DROP TABLE IF EXISTS {$tname}");
		// Make new cache table
		if( !$db->query($tsql) ) {
			print_error($db);
			exit();
		}
	}
	
	// Weblink settings table
	$weblink_settings = 
	"CREATE TABLE weblink (
		id int(11) NOT NULL AUTO_INCREMENT,
		api_key varchar(32) NOT NULL,
		rest_url varchar(255) NOT NULL,
		soap_url varchar(255) NOT NULL,
		username varchar(32) NOT NULL,
		password varchar(32) NOT NULL,
		last_cache_sweep timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
		PRIMARY KEY  (id)
	)";
	// Discard old weblink settings table
	$db->query("DROP TABLE IF EXISTS weblink");
	// Make new wordpress settings table
	if( !$db->query($weblink_settings) ) {
		print_error($db);
		exit();
	}
	
	// Outdated cache tables, if present
	$db->query("DROP TABLE IF EXISTS CachedSoapCalls");
	$db->query("DROP TABLE IF EXISTS Categories");
	$db->query("DROP TABLE IF EXISTS Courses");
	$db->query("DROP TABLE IF EXISTS Events");
	$db->query("DROP TABLE IF EXISTS Locations");
	
	// Install logs table, if not present
	$q = "CREATE TABLE IF NOT EXISTS wl_administrate_logs (
		log_id bigint(20) NOT NULL AUTO_INCREMENT,
		log_msg text NOT NULL,
		log_time bigint(20) NOT NULL,
		log_order_id int(11) NOT NULL DEFAULT '0',
		log_api_key varchar(64) NOT NULL,
		PRIMARY KEY  (log_id)
	) DEFAULT CHARSET=utf8;";
	
	if( !$db->query($q) ) {
		print_error($db);
		exit();
	}
	
	// Install orders table, if not present
	$q = "CREATE TABLE IF NOT EXISTS wl_administrate_orders (
		order_id int(11) NOT NULL AUTO_INCREMENT,
		order_status enum('I','P','C') NOT NULL DEFAULT 'I',
		order_payment_type enum('I','C') NOT NULL DEFAULT 'I',
		order_time_started bigint(20) NOT NULL,
		order_time_completed bigint(20) NOT NULL DEFAULT '0',
		order_event_id int(11) NOT NULL,
		order_session_id char(32) NOT NULL,
		order_max_step tinyint(1) NOT NULL DEFAULT '1',
		order_currency char(3) NOT NULL DEFAULT 'GBP',
		order_discount text,
		order_buyer_details text,
		order_invoice_address text,
		order_num_attendees tinyint(1) NOT NULL DEFAULT '1',
		order_attendee_details text,
		order_processor_transaction_id varchar(64) DEFAULT NULL,
		order_api_invoice_id bigint(20) NOT NULL DEFAULT '0',
		order_api_key varchar(64) NOT NULL,
		order_being_processed tinyint(1) NOT NULL DEFAULT '0',
		PRIMARY KEY  (order_id)
	) DEFAULT CHARSET=utf8;";
	
	if( !$db->query($q) ) {
		print_error($db);
		exit();
	}
}

// Copy API keys
function db_copy($db, $api_keys ) 
{
	foreach($api_keys as $key) {
		$api_key = $key["api_key"];
		$rest_url = $key["rest_url"];
		$soap_url = $key["soap_url"];
		$useranme = $key["username"];
		$password = $key["password"];
		
		$q = "
			INSERT INTO weblink (api_key, rest_url, soap_url, username, password) 
			VALUES ('{$api_key}','{$rest_url}','{$soap_url}','{$useranme}','{$password}')";
		$db->query($q);
	}
}

// Get API keys
function get_api_keys_v2($db) {
	$q = "SELECT api_key, rest_url, soap_url, username, password FROM weblink";
	$result = $db->query_fetch($q);
	
	if( !$result || (count($result)==0) ) {
		return false;
	}
	if( $db->get_last_num_records() == 1) {
		$result = array($result);
	}
	echo "API keys found in table 'weblink'.\n";	
	return $result;
}

// Get API keys
function get_api_keys_v1($db) {
	$q = "SELECT api_key, url, username, password FROM wl_api_keys";
	$result = $db->query_fetch($q);
	
	if(!$result || (count($result)==0) ) {
		return false;
	}
	if( $db->get_last_num_records() == 1) {
		$result = array($result);
	}
	
	for($i = 0; $i< count($result); $i++) {
		$soap_url = $result[$i]['url'];
		$url_parts = parse_url($soap_url);
		$rest_url = $url_parts['scheme'] . "://" . $url_parts['host'] . "/api/v2/weblink/";
		
		$result[$i]['soap_url'] = $soap_url;
		$result[$i]['rest_url'] = $rest_url;
	}
	
	echo "API keys found in table 'wl_api_keys'.\n";	
	return $result;
}

// Print usage information
function print_info() {
	echo "Usage: php weblink-components/db-install --db <weblink database> --user <username> --password <password>\n\n";
}

// Prompt user confirmation
function user_confirmation() {
	$stdin = fopen ("php://stdin","r");
	
	if( trim(fgets($stdin)) == 'yes' ) {
		return true;
	}
	echo "Aborting.\n";
	return false;
}

// Print db error
function print_error($db) {
	echo "An error has occured. Adorting.\n";
	echo $db->get_last_error()."\n\n";
}
