<?php
//  Authorize.Net Checkout Payment Processor
class AdministrateWidgetCheckoutPaymentProcessorAuthorizeNet extends AdministrateWidgetCheckoutPaymentProcessor {

	const PRODUCTION_API_URL   = 'https://secure.authorize.net/gateway/transact.dll';
	const SANDBOX_API_URL      = 'https://test.authorize.net/gateway/transact.dll';

	//  Constructor
	public function __construct(&$service, &$widget, &$order = false) {
		$this->display_name = "Authorize.Net";

		//  Call parent class
		parent::__construct($service, $widget, $order);

		// Initialize API
		//
		$config = Administrate::api()->get_payment_config('authorizenet', false);

		if( is_object($config) ) {
			$this->is_sandbox =
				(bool)($this->service->get_option('authorizenet_mode') == "sandbox");

			$this->api_url = $this->is_sandbox ?
				AdministrateWidgetCheckoutPaymentProcessorAuthorizeNet::SANDBOX_API_URL :
				AdministrateWidgetCheckoutPaymentProcessorAuthorizeNet::PRODUCTION_API_URL;

			$this->api_login_id     = $config->username;
			$this->transaction_key  = $config->password;
			$this->transaction_md5  = $config->apisignature;

			$this->is_configured = true;
		}
		else {
			$this->is_configured = false;
		}

        //  If the order isn't defined yet and an item number is in the request, initialize order
		if (!$this->order && isset($_REQUEST['item_number'])) {
			$this->set_order($this->get_order());
		}

        // RR monitor
		$this->relay_response();
	}

	//  Get the order
	public function get_order() {
		if (!$this->order || !property_exists($this, 'order')) {
			$where = array(
				'order_id'			=>	$_REQUEST['item_number'],
				'order_session_id'	=>	$_REQUEST['custom']
			);
			$this->order = new AdministrateWidgetCheckoutOrder($this->service, $where);
		}
		return $this->order;
	}
	
	// Hash function used to send and verify transactions
	private static function _transaction_hash($login_id, $sequence, $timestamp, $amount, $currency, $key)
	{
		$hash_string =  $login_id . "^" . $sequence . "^" . $timestamp . "^" . $amount . "^" . $currency;
		return ( function_exists('hash_hmac') ) ?
			hash_hmac("md5", $hash_string, $key) :
			bin2hex(mhash(MHASH_MD5, $hash_string, $key));
	}

	// Prepare parameter values to be passed on to the Authorize.net API
	private function _prepare_api_parameters()
	{
		if( !$this->is_configured ) return;

		// Transaction parameters
		//
		$rr_url = $this->widget->get_step_url()
			."api={$this->api_login_id}&step=".constant('ADM_CHECKOUT_STEP_RR')
			."&cancel_base_url=".urlencode($_REQUEST['cancel_base_url']);
		$cancel_url= preg_replace(
			'/step=[0-9]{1}/',
			'step='.constant('ADM_CHECKOUT_STEP_COURSE'),
			$_REQUEST['cancel_base_url']
		);

		$transaction = array(
			"x_login" => $this->api_login_id,
			"x_version" => "3.1",
			"x_test_request" => "false",
			"x_show_form" => "PAYMENT_FORM",

			"x_relay_response" => "true",
			"x_delim_data" => "false",
			"x_relay_url" => $rr_url,

			"x_cancel_url" => $cancel_url,
			"x_cancel_url_text" => "Cancel"
		);

		// Purchase information
		//
		$purchase = array(
			"x_method" => "CC",
			"x_po_num" => $this->order->get_session_id(),
			"x_cust_id" => $this->order->get_id(),
			'x_currency_code'   => $currency = $this->order->get_currency(),
			'x_invoice_num'     => date('YmdHis'),
		);

		if($this->order->get_event_id()) {
			$mapped_items = $this->order->get_mapped_items_and_prices();
			$item_amount = 0;
			foreach($mapped_items as $item) {
				$item_amount += $item['subtotal'] + $item['tax'] - $item['discount'];
			}
			$purchase['x_amount'] = (($this->order->get_event_gross_price() * $this->order->get_num_attendees()) + $item_amount);
			$purchase['x_description'] =
				$this->order->get_num_attendees() . $this->service->get_text(" Course Booking(s)") . ', ' . $this->order->get_course_title();

		} else {
			$track = $this->order->get_track();
			list(
				$tax_amount,
				$tax_name,
				$net_price,
				$discounted_net_price,
				$gross_price,
				$discounted_gross_price,
				) = AdministrateLearningTrackWidget::price_detail(
				$track,
				$this->service->get_option('currency', 'pricing'),
				$this->order->get_num_attendees(),
				$this->get_order()->get_discount_type(),
				$this->get_order()->get_discount_rate()
			);

			$purchase['x_amount'] = $discounted_gross_price;
			$purchase['x_description'] = $track->description;
		}

		// Random transaction sequence number, timestamp and hash
		//
		$sequence = rand(1, 1024);
		$timestamp = time();
		$hash = array(
			"x_fp_sequence"  => $sequence,
			"x_fp_timestamp" => $timestamp,
			"x_fp_hash" => AdministrateWidgetCheckoutPaymentProcessorAuthorizeNet::_transaction_hash(
					$this->api_login_id,
					$sequence,
					$timestamp,
					$purchase['x_amount'],
					$purchase['x_currency_code'],
					$this->transaction_key
				)
		);

		// Customer details
		//
		$country_id = $this->order->get_invoice_country();
		$countries = new AdministrateCountriesList();

		$customer = array(
			"x_first_name" => $this->order->get_buyer_first_name(),
			"x_last_name" => $this->order->get_buyer_last_name(),
			"x_company" => $this->order->get_buyer_company(),

			"x_address" => $this->order->get_invoice_street(),
			"x_city" => $this->order->get_invoice_city(),
			"x_zip" => $this->order->get_invoice_postal_code(),    // Postcode / ZIP
			"x_state" =>  $this->order->get_invoice_territory(),   // State / County
			"x_country" => $countries->get_name($country_id), // Country

			"x_phone" => $this->order->get_buyer_phone(),
			"x_email" => $this->order->get_buyer_email(),

			"x_ship_to_first_name" => $this->order->get_buyer_first_name(),
			"x_ship_to_last_name" => $this->order->get_buyer_last_name(),
			"x_ship_to_company" => $this->order->get_buyer_company(),
			"x_ship_to_address" => $this->order->get_invoice_street(),
			"x_ship_to_city" => $this->order->get_invoice_city(),
			"x_ship_to_zip" => $this->order->get_invoice_postal_code(),    // Postcode / ZIP
			"x_ship_to_state" =>  $this->order->get_invoice_territory(),   // State / County
			"x_ship_to_country" => $countries->get_name($country_id), // Country
		);

		// Return complete parameter list
		return array_merge($transaction,$purchase,$customer,$hash);
	}

	// Authorize.Net HTTPS Relay Response
	public function relay_response() {

		if( !$this->is_configured ) return;

		if( isset($_REQUEST["step"]) && ($_REQUEST["step"] == constant('ADM_CHECKOUT_STEP_RR')) ) {

            // Authenticate
			$md5 = $this->transaction_md5;

			if( ($md5!="") && ($md5!="disabled") ) // If not disabled in user settings
			{
				$hash_received = $_REQUEST['x_MD5_Hash'];
				$hash_str = $md5 . $_REQUEST['api'] . $_REQUEST['x_trans_id'] . $_REQUEST['x_amount'] /*. $_REQUEST['currency']*/;
				$hash_expected = strtoupper(md5($hash_str));

				if($hash_received != $hash_expected) {
					// Fraudulent transaction
					die("This transaction cannot be authenticated.");
				}
			}

			// Place the order
			$this->_place_order();

			// Show order completion page
			$redirect_url =
                preg_replace('/step=[0-9]{1}/',
                    'step='.constant('ADM_CHECKOUT_STEP_PLACE'),
                    urldecode($_REQUEST['cancel_base_url'])
                );

			echo '<script type="text/javascript">window.location.href="'.$redirect_url.'"</script>';
			echo '</head><body>';
			echo 'Thanks for your payment. <a href="'.htmlspecialchars($redirect_url).'">Click here</a> if you have not been automatically redirected.';
			echo '</body></html>';
			exit();
		}
	}

	// Place completed order
	private function _place_order() {
		$where = array(
			'order_id' => $_REQUEST['x_cust_id'],
		);
		require_once('AdministrateWidgetCheckoutOrder.php');
		$processed_order = new AdministrateWidgetCheckoutOrder($this->service, $where);

		$countries = new AdministrateCountriesList();

		$shipping = array(
				'type' => 'Shipping',
				'contact_person' => $_REQUEST['x_ship_to_first_name']. " " .$_REQUEST['x_ship_to_last_name'],
				'address1'   => $_REQUEST['x_ship_to_address'],
				'address2'   => $_REQUEST['x_ship_to_company'],
				'address3'   => $_REQUEST['x_ship_to_country'],
				'city'       => $_REQUEST['x_ship_to_city'],
				'county'     =>  $_REQUEST['x_ship_to_state'],
				'post_code'   => $_REQUEST['x_ship_to_zip'],
				'country'    => $countries->get_code($_REQUEST['x_ship_to_country']),
		);
		$billing = array(
				'type' => 'Billing',
				'contact_person' => $_REQUEST['x_first_name']. " " .$_REQUEST['x_last_name'],
				'address1'   => $_REQUEST['x_address'],
				'address2'   => $_REQUEST['x_company'],
				'address3'   => $_REQUEST['x_country'],
				'city'       => $_REQUEST['x_city'],
				'county'     =>  $_REQUEST['x_state'],
				'post_code'   => $_REQUEST['x_zip'],
				'country'    => $countries->get_code($_REQUEST['x_ship_to_country']),
		);

        $processed_order->complete(true,array($shipping,$billing));
        $this->service->db->query('UPDATE wl_administrate_orders SET order_being_processed = 0 WHERE order_id = ' . $processed_order->get_id() . ';');
	}
	
	// Is order key present in the request
	public static function has_order_key() {
		return isset( $_REQUEST['x_cust_id'] );
	}

	//  Display order form
	public function display_form() {

		if(!$this->is_configured) {
			echo "Payment integration is not configured.";
			return;
		}

		// Build API parameters
		$params = $this->_prepare_api_parameters();

		// Authorize.Net payment button
		$this->_form_end();
		$this->_form_begin($this->api_url);
		foreach( $params as $name => $val) $this->_form_parameter($name, $val);
		$this->_form_button("Place Order by Credit Card");
	}

	// Authorize.net form generation (API parameters and submit button)
	private function _form_begin($action_url) {
		echo '<form method="post" class="cc_submit_form" action="'.htmlspecialchars($action_url).'" name="pay_form" id="administrate-payment-processor-form">';
	}
	private function _form_end() {
		echo '</form>';
	}
	private function _form_parameter($name,$value){
		echo '<input type="hidden" name="'.$name.'" value="'.$value.'"/>';
	}
	private function _form_button($text) {
		echo '<input type="submit" value="'.$text.'" class="'.$this->service->add_namespace('btn', '-').'">';
	}
}
