<?php

// Define checkout steps so that we can
// view them in a textual fashion.
define('ADM_CHECKOUT_STEP_COURSE', 1);
define('ADM_CHECKOUT_STEP_ATTENDEE', 2);
define('ADM_CHECKOUT_STEP_TERMS', 3);
define('ADM_CHECKOUT_STEP_CONFIRM', 4);
define('ADM_CHECKOUT_STEP_PLACE', 5);
define('ADM_CHECKOUT_STEP_IPN', 6);		// PayPal IPN
define('ADM_CHECKOUT_STEP_RR', 7);		// Authorize.Net RR
define('ADM_CHECKOUT_BEGIN_STRIPE', 8);

$path = dirname(__FILE__);
require_once($path.'/AdministrateWidgetCheckoutPaymentProcessorPayPal.php');
require_once($path.'/AdministrateWidgetCheckoutPaymentProcessorSagePay.php');
require_once($path.'/AdministrateWidgetCheckoutPaymentProcessorAuthorizeNet.php');
require_once($path.'/AdministrateWidgetCheckoutPaymentProcessorWPM.php');

// Promo code json
add_action('send_headers', array("AdministrateWidgetCheckout", 'promo_code_json'));

//  Administrate checkout widget
class AdministrateWidgetCheckout extends WPPluginPatternWidget {

	//  Configuration
	protected $key = 'checkout';
	private $orderKey = 'order_id';
	private $eventKey = 'event_id';
	private $trackKey = 'track_id';
	private $stepKey = 'step';
	private $jsonFlag = '<!-- ADMINISTRATE_JSON_START -->';
	private $jsonFlagEnd = '<!-- ADMINISTRATE_JSON_END -->';

	private $stepFields = array(
		1	=>	array(
			'order_currency',
			'order_num_attendees',
			'order_discount',
			'order_max_step'
		),
		2	=>	array(
			'order_buyer_details',
			'order_invoice_address',
			'order_attendee_details',
			'order_max_step'
		),
		3	=>	array(
			'tac',
			'order_max_step'
		),
		4	=>	array(
			'order_max_step'
		),
		5	=>	array(
			'order_status',
			'order_payment_type',
			'order_processor_transaction_id',
			'order_max_step'
		)
	);

	/* @var $order AdministrateWidgetCheckoutOrder */
	private $order = false;

	/* @var $paymentProcessor AdministrateWidgetCheckoutPaymentProcessor */
	private $paymentProcessor = false;
	private $paymentProcessorName = false;

	/* @var $params array */
	protected $params = array();

	//  Constructor
	public function __construct(&$plugin) {

		//  Call parent constructor
		parent::__construct($plugin);

		//  Save the global variable names w/ namespacing
		$this->orderKey = $this->add_namespace($this->orderKey);
		$this->eventKey = $this->add_namespace($this->eventKey);
		$this->trackKey = $this->add_namespace($this->trackKey);
		$this->stepKey = $this->add_namespace($this->stepKey);

		//  Add the jQuery plugin & CSS
		wp_enqueue_script('jquery-ui-tabs');
		wp_enqueue_script('jquery-ui-accordion');
		wp_enqueue_script($this->add_namespace(''), $this->plugin->get_url('/js/jquery.administrate_checkout.js'), array('jquery', 'jquery-ui-tabs', 'jquery-ui-accordion'), false, true);
		wp_enqueue_script($this->add_namespace('placeholder'), $this->plugin->get_url('/js/jquery.placeholder.min.js'), array('jquery'), false, true);
		wp_enqueue_style($this->add_namespace(''), $this->_get_url('/style.css'));

		//  Start the session
		@session_start();

	}

	/**
	 * Check if a discount code is valid.
	 *
	 * @param string $code - The discount code to check.
	 * @return object/false - Object containing discount info, or false if not found.
	 */
	public static function check_discount($code, $del_count) {

		// Do a lookup against the API to find all discounts
		$discounts = Administrate::api()->raw_call('getAllDiscounts');

		$ret_discounts = array();

		// Find if we've got a valid discount
		foreach($discounts->AllDiscounts as $discount) {
			if($discount->Type == 'partner') {

				// Match on code
				if((strtolower($discount->Code) == strtolower($code) && $discount->PromoType != 'percentage_delegates') || ($discount->PromoType == 'percentage_delegates' && $del_count >= $discount->Delegates)) {

					// Found one - build an object response and return it
					$validated = new stdClass();
					$validated->type = $discount->PromoType;
					$validated->discount_code = $discount->Code;
					$validated->discount_amount = $discount->Rate;
					$ret_discounts[] = $validated;
				}
			}
		}

		// If no discount, return false
		return $ret_discounts;
	}

	/**
	 * Get discount JSON
	*/
	public static function promo_code_json()
	{
		// Check if we're doing a promo code lookup
		if(array_key_exists('validate_promo_code', $_REQUEST)) {

			// Check if promo code is valid, and return result
			$validated = AdministrateWidgetCheckout::check_discount($_REQUEST['validate_promo_code'], $_REQUEST['del_count']);

			echo json_encode($validated);
			exit;
		}
	}

	//  Run the checkout
	public function run() {
		/* This function:
		 *      Performs a pre-flight check to make darn sure there are still spaces on the event
		 *      Loads the payment processor if required
		 *      Attempts to fetch an order
		 *      If we don't have an order:
		 *          Show the events table and STOP
		 *      If we are processing an IPN response:
		 *          Process it and STOP
		 *      Validates any new information
		 *      Performs some step-specific logic, including ensuring the T's & C's were accepted
		 *      Runs the appropriate checkout step
		 */

		$this->params['currency_list'] = $this->plugin->get_option('currency_list', 'pricing');

		// If currency list was passed as string, JSON decode it
		if( is_string( $this->params['currency_list'] ) ) {
			$this->params['currency_list'] = AdministrateEventWidget2::decode_currency_list( $this->params['currency_list'] );
		}

		// Currency list empty or not parsing
		if( !is_array($this->params['currency_list']) || empty($this->params['currency_list']) ) {
			$this->params['currency_list'] = false;
		}

		// If we've got an event to work with, pull a fresh copy of it from the DB and check if we actually have spaces
		if(array_key_exists($this->eventKey, $_REQUEST) && $_REQUEST[$this->eventKey] > 0) {
			// Instruct API to fetch fresh event, without caching
			$pre_flight = Administrate::api()->get_event($_REQUEST[$this->eventKey], false);

			// check if the event is sold out, honouring the setting to check 'places'
			if($pre_flight->is_sold_out()) {
				return $this->plugin->display_widget('event_table', array('errors' => 'Unfortunately, the course you have selected is now sold out. Please choose another event.'));
			}
		}

		// Include order and log DAO's
		require_once($this->get_path('/AdministrateWidgetCheckoutOrder.php'));
		require_once($this->get_path('/AdministrateWidgetCheckoutLog.php'));

		$log = new AdministrateWidgetCheckoutLog($this->plugin);

		//  Figure out what step we're on
		if (!isset($_REQUEST[$this->stepKey]) || !is_numeric($_REQUEST[$this->stepKey])) {
			$current_step = constant('ADM_CHECKOUT_STEP_COURSE');
		} else {
			$current_step = intval($_REQUEST[$this->stepKey]);
		}

		// Get order currency
		//
		if( isset($_SESSION[$this->orderKey]) ) {
			$order_id = $_SESSION[$this->orderKey];
		}
		else if( array_key_exists($this->orderKey, $this->params) && $this->params[$this->orderKey] ) {
			$order_id = $this->params[$this->orderKey];
		}
		else {
			$order_id = false;
		}

		if( $order_id !== false ) {
			$previous_order = new AdministrateWidgetCheckoutOrder(
				$this->plugin,
				array( $this->strip_namespace($this->orderKey) => $order_id )
			);
			$order_currency = $previous_order->get_currency();
		}
		else {
			if( array_key_exists('event_currency', $_REQUEST) ) {
				$order_currency = $_REQUEST['event_currency'];
			}
			elseif( !array_key_exists('currency', $this->params) || !$this->params['currency'] ) {
				$order_currency = $this->plugin->get_option('currency', 'pricing');
			}
			// Have a last-gasp attempt at grabbing the currency from the paypal IPN post
			elseif(array_key_exists('mc_currency', $_REQUEST)) {
				$order_currency = $_REQUEST['mc_currency'];
			}
		}

		// If checkout for the same event in a different,
		//
		if( array_key_exists('event_currency', $_REQUEST) && array_key_exists('event_currency', $this->params) ) {
			$this->params['currency'] = $this->params['event_currency'];

			if( $order_currency!=$_REQUEST['event_currency'] ) {
				// Reset the order
				unset($_SESSION[$this->orderKey]);
				$this->params[$this->orderKey] = false;
			}
		}

		// If we're enabling pay by CC and we're passed the terms step
		if( intval($this->get_option('payment_by_cc')) ) {

			if( array_key_exists('currency_list',$this->params) && is_array($this->params['currency_list']) ) {

				// Get processor associated with currency
				$processor = $this->params['currency_list'][$order_currency];
			}
			else {
				// Default processor
				$processor = $this->plugin->get_option('payment_processor', 'checkout');
			}

			// Now load the class specific to the processor we're using
			$class = __CLASS__ . 'PaymentProcessor' . $processor;

			if( !class_exists($class) ) {
				$processor = "PayPal";
				$class = __CLASS__ . 'PaymentProcessorPayPal';
			}

			$this->paymentProcessorName = $processor;

			if( $current_step >= constant('ADM_CHECKOUT_STEP_CONFIRM') ) {
				$this->paymentProcessor = new $class($this->plugin, $this);
				if ($this->paymentProcessor->order && $this->paymentProcessor->order->exists()) {
					$this->order = $this->paymentProcessor->get_order();
				}
			}
		}

		// Try and fetch order
		if(!$this->order) {

			// If a zero event ID or track ID was passed, reset the session
			if(
				(isset($_REQUEST[$this->eventKey]) && (intval($_REQUEST[$this->eventKey]) == 0))
				|| (isset($_REQUEST[$this->trackKey]) && (intval($_REQUEST[$this->trackKey]) == 0))
			) {
				$this->order = false;
				if(isset($_SESSION[$this->orderKey])) {
					unset($_SESSION[$this->orderKey]);
				}

			// Otherwise try to bring up the order in the session
			} else {
				// Take order key from session if we can, otherwise check if it's set in postdata
				$where = array();
				if(isset($_SESSION[$this->orderKey])) {
					$where[$this->strip_namespace($this->orderKey)] = $_SESSION[$this->orderKey];
				} elseif(isset($_POST[$this->orderKey])) {
					$where[$this->strip_namespace($this->orderKey)] = $_POST[$this->orderKey];
				} elseif(isset($_POST['item_number'])) {
					// This handles multiple requests from Paypal's Return to Merchant button
					$where['order_id'] = $_POST['item_number'];
				}

				// If we have anything to query for the order, query for it now
				if(!empty($where)) {
					$this->order = new AdministrateWidgetCheckoutOrder($this->plugin, $where);
					if (!$this->order->exists()) {
						$this->order = false;
					}
				}

				// Was an event ID passed in?
				if(
					(isset($_REQUEST[$this->eventKey]) && (intval($_REQUEST[$this->eventKey]) > 0))
					|| (isset($_REQUEST[$this->trackKey]) && (intval($_REQUEST[$this->trackKey]) > 0))
				) {

					// If we've not got an order, or the order we've got doesn't match the requested key
					if(
						!$this->order
						|| (isset($_REQUEST[$this->eventKey]) && ($this->order->get_event_id() != $_REQUEST[$this->eventKey]))
						|| (isset($_REQUEST[$this->trackKey]) && ($this->order->get_track_id() != $_REQUEST[$this->trackKey]))
					) {

						// Build a new order
						$this->order = new AdministrateWidgetCheckoutOrder($this->plugin);

						// Use post/get currency, or read from settings
						if( array_key_exists('event_currency', $_REQUEST) ) {
							$this->params['currency'] = $_REQUEST['event_currency'];
						}
						else if(isset($_REQUEST['order_currency'])) {
							$this->params['currency'] = $_REQUEST['order_currency'];
						}
						else {
							$this->params['currency'] = $this->plugin->get_option('currency', 'pricing');
						}

						// Build the new order and assign to the session
						if($_REQUEST[$this->eventKey]) {
							$event = Administrate::api()->get_event($_REQUEST[$this->eventKey]);

							// If the event is not available in the desired currency,
							if( ! $event->get_price_by_currency( $this->params['currency'] ) ) {
								// Start the order in the event default currency
								$this->params['currency'] = $event->get_default_currency();
							}
							$this->order->create($_REQUEST[$this->eventKey], session_id(), $this->params['currency']);
						} else {
							$this->order->create(null, session_id(), $this->params['currency'], $_REQUEST[$this->trackKey]);
						}
						$_SESSION[$this->orderKey] = $this->order->get_id();
					}
				}
			}
		}

		// If we *still* don't have an order - just show the event table
		if( !$this->order ) {
			return $this->plugin->display_widget('event_table');
		}

		// If we have item id's set for mapped items, save them using our setter then nuke them from the POST
		// so that they don't overwrite themselves with a serialized out PHP Array() later.
		if(array_key_exists('order_item_ids', $_POST)) {
			$this->order->set_item_ids($_POST['order_item_ids']);
			unset($_POST['order_item_ids']);
		} else {
			$this->order->set_item_ids(array());
		}

		$log->create('Found order, beginning step ' . ($current_step-1) . '.', $this->order->get_id());

		// If this is an IPN response, handle the payment and stop
		// Further, we've had an IPN which has marked this order as completed, don't
		// try and continue processing it.
		if( ($current_step == constant('ADM_CHECKOUT_STEP_IPN')) && !$this->order->is_complete() ) {
			$log->create('Received IPN response, as step is 6 and order is not complete.', $this->order->get_id());
			$this->order->receive_ipn();
			return;
		}

		// At this point, we must have an order - rejoice!
		// Read and use some defaults from the options.
		$this->_set_options();
		$this->widgetTitle = $this->_generate_title();

		// Default $errors to empty array, we'll populate it if we have any.
		$errors = array();

		/* Max step is the highest step that we've completed and
		 * validated - this determines what steps are available
		 * in the checkout widget.
		 */
		if($current_step > $this->order->get_max_step()) {

			// Set it in the post, so that we can save it with
			// the rest of the post data from the order.
			$_POST['order_max_step'] = $current_step;
		}

		if($current_step > constant('ADM_CHECKOUT_STEP_COURSE')) {
			$errors = $this->order->save($this->_prepare_api_fields($current_step - 1, $_POST));
		}

		// #### EXTRA VALIDATION FOR SPECIFIC STEPS ####

		// If we're in an order and there's no currency, reset to first step
		$currency = $this->order->get_currency();
		if($current_step > constant('ADM_CHECKOUT_STEP_COURSE') && empty($currency)) {
			$log->create('Tried to progress past step course with no currency set, resetting to step 1.', $this->order->get_id());
			$current_step = 1;
		}

		// If we're on the confirmation step and the T's & C's
		// weren't accepted, add the validation error.
		if($current_step == constant('ADM_CHECKOUT_STEP_CONFIRM') && !isset($_REQUEST['tac']) && !isset($_REQUEST[$this->plugin->add_namespace(array('processor', 'cancel'))])) {
			$log->create('Tried to progress past T&C step without ticking checkbox.', $this->order->get_id());
			$errors[] = __('You must accept the Terms &amp; Conditions before proceeding.', 'administrate');
		}

		// This field is hidden on the input form, so if it's filled in, a bot did it.
		if(!empty($_REQUEST['validation_code'])) {
			$errors[] = __('Hidden field populated, suspecting bot activity and not proceeding with checkout.', 'administrate');
		}

		$log->create('Completed logic checks for step ' . ($current_step-1) . '.', $this->order->get_id());
		// If we've errors revert to previous step to allow correction.
		if($current_step > constant('ADM_CHECKOUT_STEP_COURSE') && !empty($errors)) {
			$current_step--;
		}

		// Form action is the 'next' step - e.g. where the 'submit' button goes.
		$form_action = $this->get_step_url($current_step + 1);

		// Run the appropriate step
		$step_content = array();
		$complete_content = false;
		if($current_step == constant('ADM_CHECKOUT_STEP_PLACE')) {

			// _run_completion will place the order with Administrate
			$complete_content = $this->_run_completion($errors);
		} elseif($current_step == constant('ADM_CHECKOUT_BEGIN_STRIPE')) {
			$payment_processor = $this->paymentProcessorName;
			if($payment_processor == 'Stripe') {
				// If paying by invoice
				if(!Administrate::get_is_invoice_payment_from_post() && $this->paymentProcessor->get_token()) {
					// this is the response from the Stripe server, just return it as this should be an AJAX call
                    echo $this->jsonFlag;
					echo $this->paymentProcessor->card_transaction();
					echo $this->jsonFlagEnd;
					exit;
				}
			}
		} else {
			if(
				(!isset($_REQUEST['ajax']) && ($current_step >= constant('ADM_CHECKOUT_STEP_COURSE')))
				|| ($current_step == constant('ADM_CHECKOUT_STEP_COURSE'))
			) {
				$step_content[constant('ADM_CHECKOUT_STEP_COURSE')] = $this->_run_step1($errors);
			}

			if(
				(!isset($_REQUEST['ajax']) && ($current_step >= constant('ADM_CHECKOUT_STEP_ATTENDEE')))
				|| ($current_step == constant('ADM_CHECKOUT_STEP_ATTENDEE'))
			) {
				$step_content[constant('ADM_CHECKOUT_STEP_ATTENDEE')] = $this->_run_step2($errors);
			}

			if(
				(!isset($_REQUEST['ajax']) && ($current_step >= constant('ADM_CHECKOUT_STEP_TERMS')))
				|| ($current_step == constant('ADM_CHECKOUT_STEP_TERMS'))
			) {
				$step_content[constant('ADM_CHECKOUT_STEP_TERMS')] = $this->_run_step3($errors);
			}

			if(
				(!isset($_REQUEST['ajax']) && ($current_step >= constant('ADM_CHECKOUT_STEP_CONFIRM')))
				|| ($current_step == constant('ADM_CHECKOUT_STEP_CONFIRM'))
			) {
				$step_content[constant('ADM_CHECKOUT_STEP_CONFIRM')] = $this->_run_step4($errors);
			}
		}

		// If the request was made via AJAX, just spit out the requested step and exit
		if(isset($_REQUEST['ajax']) && ($current_step < constant('ADM_CHECKOUT_STEP_PLACE'))) {

			$log->create('Echoing out json for step ' . ($current_step) . ' of checkout.', $this->order->get_id());
			echo $this->jsonFlag;
			$json = array(
				"step"		=>	$current_step,
				"content"	=>	$step_content[$current_step]
			);
			echo json_encode($json);
			echo $this->jsonFlagEnd;
			exit;

		}
		// Otherwise show the whole checkout interface
		else {

			$log->create('Rendering out complete form for ' . ($current_step) . ' of checkout.', $this->order->get_id());
			// If there is content to display, display it in the checkout template
			if ((count($step_content) > 0) || $complete_content) {

				$payment_by_cc = $this->get_option('payment_by_cc');

				if($this->order->get_event_id()) {
					$is_event_free = ($this->order->get_event_price() ? $this->order->get_event_price()->is_zero() : true);
				}

				// Include the template
				ob_start();
				require_once($this->get_path('/views/checkout.php'));
				return ob_get_clean();

			}
		}
	}

	//  Run step 1
	private function _run_step1($errors = array()) {
		// Refresh cache for the event before proceeding with event booking
		if($this->order->get_event_id()) {
			$this->order->fetch_event(true);

			$possible_mapped_items = Administrate::api()->get_mapped_items(
				$this->order->get_event()->get_course()->get_id(),
				$this->order->get_event()->get_location_id()
			);

			//  Includes the step 1 template
			$content = $this->get_include_contents(
				$this->get_path('/views/checkout_step1.php'),
				array(
					'course' => $this->widgetTitle,
					'course_fields' => $this->plugin->get_event_fields($this->order->get_event(), $this->get_option('step1_course_fields')),
					'num_attendees' => $this->order->get_num_attendees(),
					'promo_code' => $this->order->get_promo_code(),
					'prices' => $this->order->get_event_prices(),
					'pricing_basis' => $this->plugin->get_option('basis', 'pricing'),
					'tax_label' => $this->plugin->get_option('tax_label', 'pricing'),
					'prices_include_taxes' => $this->plugin->to_boolean($this->plugin->get_option('inc_taxes', 'pricing')),
					'default_currency' => $this->order->get_currency(),
					'show_currency_indicator' => $this->plugin->get_option('show_currency_indicator', 'pricing'),
					'currency_indicator' => $this->plugin->get_option('currency_indicator', 'pricing'),
					'event' => $this->order->get_event(),
					'errors' => $errors,
					'show_promo_codes' => $this->plugin->get_option('show_promo_codes', 'checkout'),
					'is_event_free' => ($this->order->get_event_price() ? $this->order->get_event_price()->is_zero() : true),
					'show_sessions' => $this->plugin->get_bool_option('show_sessions', 'event'),
					'show_session_titles' => $this->plugin->get_bool_option('show_session_titles', 'event'),
					'show_times' => $this->plugin->get_bool_option('show_times', 'event'),
					'possible_mapped_items' => $possible_mapped_items,
					'item_ids' => $this->order->get_item_ids()
				)
			);
		} else {
			$content = $this->get_include_contents(
				$this->get_path('/views/checkout_step1_track.php'),
				array(
					'track' => $this->order->get_track(),
					'num_attendees' => $this->order->get_num_attendees(),
					'promo_code' => $this->order->get_promo_code(),
					'pricing_basis' => $this->plugin->get_option('basis', 'pricing'),
					'show_currency_indicator' => $this->plugin->get_option('show_currency_indicator', 'pricing'),
					'currency' => $this->order->get_currency(),
					'errors' => $errors,
				)
			);
		}

		//  Return the content to display
		return $content;

	}

	//  Run step 2
	private function _run_step2($errors = array()) {

		//  Save reference to table fields
		$tableFields = $this->plugin->get_table_fields('orders');

		//  Set the buyer fields
		$buyerValues = $this->_get_serialized_field_values($tableFields['order_buyer_details']['serialize'], $this->order->get_buyer_details(), 'order_buyer_details_');

		//  Set the invoice address fields
		$invoiceValues = $this->_get_serialized_field_values($tableFields['order_invoice_address']['serialize'], $this->order->get_invoice_address(), 'order_invoice_address_');
		if (empty($invoiceValues['country'])) {
			$invoiceValues['country'] = $this->plugin->get_option('default_country', 'checkout');
		}

		//  Set the attendee fields
		$attendeeValues = $this->_get_serialized_field_values($tableFields['order_attendee_details']['serialize'], $this->order->get_attendee_details(), 'order_attendee_details_', $this->order->get_num_attendees());

		$is_event_free = false;
		if($this->order->get_event_id()) {
			$is_event_free = $this->order->get_event_price()->is_zero();
		} else {
			$track_price = AdministrateLearningTrackWidget::get_price($this->order->get_track());
			$is_event_free = ($track_price->amount == 0);
		}

		//  Includes the step 1 template
		$content = $this->get_include_contents(
			$this->get_path('/views/checkout_step2.php'),
			array(
				'buyer_fields'			=>	$tableFields['order_buyer_details']['serialize'],
				'buyer_values'			=>	$buyerValues,
				'invoice_fields'		=>	$tableFields['order_invoice_address']['serialize'],
				'invoice_values'		=>	$invoiceValues,
				'attendee_fields'		=>	$tableFields['order_attendee_details']['serialize'],
				'attendee_values'		=>	$attendeeValues,
				'num_attendees'			=>	$this->order->get_num_attendees(),
				'countries'				=>	$this->get_countries(),
				'errors'				=>	$errors,
				'show_notes'			=>	$this->plugin->get_option('show_notes_fields', 'checkout'),
				'default_order_note'	=>	$this->plugin->get_option('default_order_note', 'checkout'),
				'default_student_note'	=>	$this->plugin->get_option('default_student_note', 'checkout'),
				'is_event_free'			=>	$is_event_free
			)
		);

		//  Return the content to display
		return $content;

	}

	//  Run step 3
	private function _run_step3($errors = array()) {

		//  Includes the step 3 template
		$content = $this->get_include_contents(
			$this->get_path('/views/checkout_step3.php'),
			array(
				'errors'	=>	$errors
			)
		);

		//  Return the content to display
		return $content;

	}

	//  Run step 4
	private function _run_step4($errors = array()) {

		//  Save reference to table fields
		$tableFields = $this->plugin->get_table_fields('orders');

		//  Set the buyer fields
		$buyerValues = $this->_get_serialized_field_values($tableFields['order_buyer_details']['serialize'], $this->order->get_buyer_details(), 'order_buyer_details_');

		//  Set the invoice address fields
		$invoiceValues = $this->_get_serialized_field_values($tableFields['order_invoice_address']['serialize'], $this->order->get_invoice_address(), 'order_invoice_address_');

		//  Set the attendee fields
		$attendeeValues = $this->_get_serialized_field_values($tableFields['order_attendee_details']['serialize'], $this->order->get_attendee_details(), 'order_attendee_details_', $this->order->get_num_attendees());

		//  If there is a payment processor, set the order to it
		if ($this->paymentProcessor) {
			$this->paymentProcessor->set_order($this->order);
		}

		if($this->order->get_event_id()) {
			$event = $this->order->get_event();

			// Tax label
			$tax_label_custom = $event->get_tax_name();
			$tax_label = $event->is_eu_digital_good() && !empty($tax_label_custom) ?
				$tax_label_custom : $this->plugin->get_option('tax_label', 'pricing');

			//  Includes the step 4 template
			$content = $this->get_include_contents(
				$this->get_path('/views/checkout_step4.php'),
				array(
					'buyer_fields' => $tableFields['order_buyer_details']['serialize'],
					'buyer_values' => $buyerValues,
					'invoice_fields' => $tableFields['order_invoice_address']['serialize'],
					'invoice_values' => $invoiceValues,
					'attendee_fields' => $tableFields['order_attendee_details']['serialize'],
					'attendee_values' => $attendeeValues,
					'num_attendees' => $this->order->get_num_attendees(),
					'course_fields' => $this->plugin->get_event_fields($this->order->get_event(), $this->get_option('step4_course_fields')),
					'order_id' => $this->order->get_id(),
					'billing_address' => $this->get_option('billing_address'),
					'course' => $this->widgetTitle,
					'prices' => $this->order->get_event_prices(),
					'pricing_basis' => $this->plugin->get_option('basis', 'pricing'),
					'tax_label' => $tax_label,
					'prices_include_taxes' => $this->plugin->to_boolean($this->plugin->get_option('inc_taxes', 'pricing')),
					'default_currency' => $this->order->get_currency(),
					'show_currency_indicator' => $this->plugin->get_option('show_currency_indicator', 'pricing'),
					'currency_indicator' => $this->plugin->get_option('currency_indicator', 'pricing'),
					'event' => $event,
					'errors' => $errors,
					'show_notes' => $this->plugin->get_option('show_notes_fields', 'checkout'),
					'is_event_free' => ($this->order->get_event_price() ? $this->order->get_event_price()->is_zero() : true),
					'mapped_items' => $this->order->get_mapped_items_and_prices()
				)
			);
		} else {
			$content = $this->get_include_contents(
				$this->get_path('/views/checkout_step4_track.php'),
				array(
					'track' => $this->order->get_track(),
					'num_attendees' => $this->order->get_num_attendees(),
					'promo_code' => $this->order->get_promo_code(),
					'pricing_basis' => $this->plugin->get_option('basis', 'pricing'),
					'show_currency_indicator' => $this->plugin->get_option('show_currency_indicator', 'pricing'),
					'currency' => $this->order->get_currency(),
					'buyer_fields' => $tableFields['order_buyer_details']['serialize'],
					'buyer_values' => $buyerValues,
					'invoice_fields' => $tableFields['order_invoice_address']['serialize'],
					'invoice_values' => $invoiceValues,
					'attendee_fields' => $tableFields['order_attendee_details']['serialize'],
					'attendee_values' => $attendeeValues,
					'errors' => $errors,
				)
			);
		}
		//  Return the content to display
		return $content;

	}

	//  Run step 5 (ie, completion)
	private function _run_completion($errors) {
		$payment_processor = $this->paymentProcessorName;
		$order_status_ok = true;

		if($payment_processor == 'Stripe') {

			// If paying by invoice
			if(Administrate::get_is_invoice_payment_from_post()) {
				$order_status_ok = true;
			} else {
				// Error handling is now handled by Stripe on the client side.
				$order_status_ok = true;
				$this->order->complete(true);
			}
		}

		// If the order isn't complete, process it
		if(!$this->order->is_complete() && $order_status_ok) {

			/* If we're doing a SagePay order, there is no differentiation
			 * between SuccessURL and IPN - it only accepts one URL. For this
			 * reason we ensure that the order is processed and not submitted
			 * as 'pending.'
			 */
			$skip_pending = ($payment_processor == 'SagePay') || isset($_GET['crypt']);
			$order_status_ok = $this->order->complete($skip_pending);
		}

		//  Kill the session variable
		unset($_SESSION[$this->orderKey]);

		if( is_object($this->paymentProcessor) && property_exists($this->paymentProcessor, "display_name") ) {
			$processor_display_name = $this->paymentProcessor->display_name;
		}
		else {
			$currency_list  = $this->params['currency_list'];

			if( array_key_exists('event_currency', $_REQUEST) && is_array($currency_list) ) {
				$event_currency = $_REQUEST['event_currency'];

				if( array_key_exists( $event_currency, $currency_list) ) {
					$processor_display_name = $currency_list[ $event_currency ];
				}
			}
			else {
				$processor_display_name = $payment_processor;
			}
		}

		//  Includes the step 5 template
		$content = $this->get_include_contents(
			$this->get_path('/views/checkout_complete.php'),
			array(
				'payment_type'		=>	$this->order->get_payment_type(),
				'payment_processor'	=>	$processor_display_name,
				'errors'			=>	$errors,
				'order_status_ok' => $order_status_ok,
				'order_num'			=>	$this->order->get_api_invoice_id(),
				'order_currency'			=>	$this->order->get_currency()
			)
		);

		//  Return the content to display
		return $content;

	}

	//  Prepare fields
	private function _prepare_api_fields($step, $inputFields) {
		$outputFields = array();
		foreach ($inputFields as $key=>$value) {
			if (isset($this->stepFields[$step])) {
				foreach ($this->stepFields[$step] as $allowedField) {
					if (substr($key, 0, strlen($allowedField)) == $allowedField) {
						$outputFields[$key] = $value;
					}
				}
			}
		}
		return $outputFields;
	}

	//  Get world countries
	public function get_countries() {
		$countries = new AdministrateCountriesList();
		return $countries->get_list();
	}

	//  Validate the payment
	private function _validate_payment() {

		//  Initialize errors array
		$errors = array();

		//  If the payment type is credit card, validate using processor specific routines
		if (!isset($_REQUEST['order_payment_type'])) {
			$paymentError = $this->paymentProcessor->validate_payment();
			if ($paymentError) {
				$errors['payment'] = $paymentError;
			}
		}

		//  Return the errors
		return $errors;

	}

	//  Process the payment
	private function _process_payment($txnId=false) {
		return $this->paymentProcessor->process_payment();
	}

	//  Get field values for a serialized set of fields
	private function _get_serialized_field_values($fields, $savedValues, $fieldPrefix, $num = false) {
		$savedValues = unserialize($savedValues);
		$values = array();
		foreach ($fields as $field) {
			if (isset($_POST[$fieldPrefix.$field])) {
				$values[$field] = $_POST[$fieldPrefix.$field];
			} else if (is_array($savedValues) && isset($savedValues[$field])) {
				$values[$field] = $savedValues[$field];
			} else {
				if ($num) {
					$values[$field] = array();
					for ($i = 1; $i <= $num; ++$i) {
						array_push($values[$field], '');
					}
				} else {
					$values[$field] = '';
				}
			}
		}
		return $values;
	}

	//  Generate the widget title
	private function _generate_title() {
		if($this->order->get_event_id()) {
			$title = $this->order->get_course_title() . '<br>';

			$show_classroom_dates = $this->plugin->get_bool_option('show_classroom_dates', 'event');
			$date_range = $this->order->get_event_dates($show_classroom_dates);
			$online_event = $this->order->get_event()->is_online();
			if ($this->order->get_event()->is_persistent()) {
				$title .= 'Access Duration: ' . $this->order->get_event()->get_access_duration() . ' days.';
			} else {
				$title .= $this->service->t_dates($date_range, $online_event, $this->order->get_event()->is_persistent());
			}

			if ($this->plugin->get_option('show_times', 'event')) {
				$title .= '<br>' . $this->plugin->t_times($this->order->get_event_times());
			}
		} else {
			$title = $this->order->get_track()->name . '<br>';
		}
		return $title;
	}

	//  Get the step URL
	public function get_step_url($num, $includeHost = false) {
		$href = $this->get_base_url($includeHost);
		if (strpos($href, '?') > -1) {
			$href .= '&';
		} else {
			$href .= '?';
		}
		$href .= $this->stepKey . '=' . $num;
		return $href;
	}

	//  Get the base URL
	public function get_base_url($includeHost = false) {

		//  If the base URL isn't set yet, set it now
		if (!property_exists($this, 'baseUrl')) {
			$params = $_GET;
			if (isset($params[$this->stepKey])) {
				unset($params[$this->stepKey]);
			}
			$this->baseUrl = parse_url($_SERVER['REQUEST_URI'], PHP_URL_PATH);
			if (!empty($params)) {
				$this->baseUrl .= '?' . http_build_query($params);
			}
		}

		//  If we're including the host ...
		if ($includeHost) {

			//  If the full base URL isn't set, set it now
			if (!property_exists($this, 'baseUrlFull')) {
				$tmp = 'http';

				// HTTPS can be off if the var is not set (most common) by
				// also if it's ACTUALLY set to 'off.' Handle both cases.
				if (array_key_exists('HTTPS', $_SERVER) && $_SERVER['HTTPS'] !== 'off') {
					$tmp .= 's';
				}
				$tmp .= '://'.$_SERVER['SERVER_NAME'];
				if (($_SERVER["SERVER_PORT"] != 80) && ($_SERVER['SERVER_PORT'] != 443)) {
					$tmp .= ':'.$_SERVER['SERVER_PORT'];
				}
				$this->baseUrlFull = $tmp . $this->baseUrl;
			}

			//  Return the full base URL
			return $this->baseUrlFull;

		//  Or else return the base URL
		} else {
			return $this->baseUrl;
		}

	}

	//  Get the reset URL
	private function _get_reset_url() {
		$url = $this->get_base_url();
		if (empty($_SERVER['QUERY_STRING'])) {
			$url .= '?';
		} else {
			$url .= '&';
		}
		$url .= $this->eventKey . '=0';
		return $url;
	}

	//  Display the payment form
	public function display_payment_form() {
		$this->paymentProcessor->display_form();
	}

	//  Get the order
	public function get_order() {
		return $this->order;
	}

	//  Get the order event
	public function get_order_event() {
		return $this->order->get_event();
	}
}
